#include "dxut.h"
#include "BulletUtilities.h"
#include "Vertex.h"

#include "BulletCollision/CollisionShapes/btShapeHull.h"

D3DXMATRIX ConvertMatrix( btTransform &trn ) 
{
	btVector3 R = trn.getBasis().getColumn(0);
	btVector3 U = trn.getBasis().getColumn(1);
	btVector3 L = trn.getBasis().getColumn(2);
	btVector3 P = trn.getOrigin();

	D3DXVECTOR3 vR, vU, vL, vP;
	vR.x = R.x(); vR.y = R.y(); vR.z = R.z();
	vU.x = U.x(); vU.y = U.y(); vU.z = U.z();
	vL.x = L.x(); vL.y = L.y(); vL.z = L.z();
	vP.x = P.x(); vP.y = P.y(); vP.z = P.z();

	D3DXMATRIX matOutput;
	matOutput._11 = vR.x; matOutput._12 = vR.y; matOutput._13 = vR.z; matOutput._14 = 0.f;
	matOutput._21 = vU.x; matOutput._22 = vU.y; matOutput._23 = vU.z; matOutput._24 = 0.f;
	matOutput._31 = vL.x; matOutput._32 = vL.y; matOutput._33 = vL.z; matOutput._34 = 0.f;
	matOutput._41 = vP.x; matOutput._42 = vP.y; matOutput._43 = vP.z; matOutput._44 = 1.f;

	return matOutput;
}

D3DXVECTOR3 ConvertVector( btVector3 &vec )
{
	return D3DXVECTOR3( vec.getX(), vec.getY(), vec.getZ() );
}

btVector3 ConvertVector( const D3DXVECTOR3 *vec )
{
	return btVector3(vec->x, vec->y, vec->z );
}

btVector3 ConvertVector( D3DXVECTOR3 &vec )
{
	return btVector3(vec.x, vec.y, vec.z );
}

const btVector3 rotate( const btQuaternion& quat, const btVector3 & vec )
{
    float tmpX, tmpY, tmpZ, tmpW;
    tmpX = ( ( ( quat.getW() * vec.getX() ) + ( quat.getY() * vec.getZ() ) ) - ( quat.getZ() * vec.getY() ) );
    tmpY = ( ( ( quat.getW() * vec.getY() ) + ( quat.getZ() * vec.getX() ) ) - ( quat.getX() * vec.getZ() ) );
    tmpZ = ( ( ( quat.getW() * vec.getZ() ) + ( quat.getX() * vec.getY() ) ) - ( quat.getY() * vec.getX() ) );
    tmpW = ( ( ( quat.getX() * vec.getX() ) + ( quat.getY() * vec.getY() ) ) + ( quat.getZ() * vec.getZ() ) );
    return btVector3(
        ( ( ( ( tmpW * quat.getX() ) + ( tmpX * quat.getW() ) ) - ( tmpY * quat.getZ() ) ) + ( tmpZ * quat.getY() ) ),
        ( ( ( ( tmpW * quat.getY() ) + ( tmpY * quat.getW() ) ) - ( tmpZ * quat.getX() ) ) + ( tmpX * quat.getZ() ) ),
        ( ( ( ( tmpW * quat.getZ() ) + ( tmpZ * quat.getW() ) ) - ( tmpX * quat.getY() ) ) + ( tmpY * quat.getX() ) )
    );
}

btCollisionShape * CreateTowerCircle(btDynamicsWorld *pWorld, const btVector3& offsetPosition, int stackSize, int rotSize, const btVector3& boxSize)
{
	btBoxShape* blockShape = new btBoxShape(btVector3(boxSize[0]-COLLISION_RADIUS,boxSize[1]-COLLISION_RADIUS,boxSize[2]-COLLISION_RADIUS));

	btTransform trans;
	trans.setIdentity();

	float mass = 1.f;
	btVector3 localInertia(0,0,0);
	blockShape->calculateLocalInertia(mass,localInertia);

	float radius = 1.4f * rotSize * boxSize[0] / SIMD_PI;

	// create active boxes
	btQuaternion rotY(0,1,0,0);
	float posY = boxSize[1];

	for(int i=0;i<stackSize;i++) 
	{
		for(int j=0;j<rotSize;j++) 
		{
			trans.setOrigin(offsetPosition+  rotate(rotY,btVector3(0.0f , posY, radius)));
			trans.setRotation(rotY);
			btRigidBody *body = CreateRigidBody(mass,trans,blockShape);			
			pWorld->addRigidBody(body );		

			rotY *= btQuaternion(btVector3(0,1,0),SIMD_PI/(rotSize*0.5f));
		}

		posY += boxSize[1] * 2.0f;
		rotY *= btQuaternion(btVector3(0,1,0),SIMD_PI/(float)rotSize);
	}	

	return blockShape;
}
btCollisionShape * CreateWall(btDynamicsWorld *world, const btVector3& offsetPosition,int stackSize,const btVector3& boxSize)
{	
	btBoxShape* blockShape = new btBoxShape(btVector3(boxSize[0]-COLLISION_RADIUS,boxSize[1]-COLLISION_RADIUS,boxSize[2]-COLLISION_RADIUS));

	float mass = 1.f;
	btVector3 localInertia(0,0,0);
	blockShape->calculateLocalInertia(mass,localInertia);

	float diffX = boxSize[0] * 1.0f;
	float diffY = boxSize[1] * 1.0f;
	float diffZ = boxSize[2] * 1.0f;

	float offset = -stackSize * (diffZ * 2.0f) * 0.5f;
	btVector3 pos(0.0f, diffY, 0.0f);

	btTransform trans;
	trans.setIdentity();

	while(stackSize) 
	{
		for(int i=0;i<stackSize;i++) 
		{
			pos[2] = offset + (float)i * (diffZ * 2.0f);

			trans.setOrigin(offsetPosition + pos);
			btRigidBody *body = CreateRigidBody(mass,trans,blockShape);			
			world->addRigidBody(body );				
		}
		offset += diffZ;
		pos[1] += (diffY * 2.0f);
		stackSize--;
	}

	return blockShape;
}

btCollisionShape * CreatePyramid(btDynamicsWorld *world, const btVector3& offsetPosition,int stackSize, const btVector3& shapeSize, int shapeType)
{	
	float space = 0.0f;
	btVector3 pos(0.0f, shapeSize[1], 0.0f);

	btCollisionShape *collisionShape = NULL;
	if( shapeType == BOX_SHAPE_PROXYTYPE )
	{
		collisionShape = new btBoxShape(btVector3(shapeSize[0]-COLLISION_RADIUS,shapeSize[1]-COLLISION_RADIUS,shapeSize[2]-COLLISION_RADIUS));
	}
	else if( shapeType == CYLINDER_SHAPE_PROXYTYPE )
	{
		collisionShape = new btCylinderShape( btVector3(shapeSize[0]-COLLISION_RADIUS,shapeSize[1]-COLLISION_RADIUS,shapeSize[2]-COLLISION_RADIUS) );	
	}
	else
	{
		assert( "Invalid shape type on call to CreatePyramid()" );
	}
	
	btTransform trans;
	trans.setIdentity();

	float mass = 1.f;
	btVector3 localInertia(0,0,0);
	collisionShape->calculateLocalInertia(mass,localInertia);
	
	float diffX = shapeSize[0];
	float diffY = shapeSize[1];
	float diffZ = shapeSize[2];
	
	float offsetX = -stackSize * (diffX * 2.0f + space) * 0.5f;
	float offsetZ = -stackSize * (diffZ * 2.0f + space) * 0.5f;
	while(stackSize) 
	{
		for(int j=0;j<stackSize;j++) 
		{
			pos[2] = offsetZ + (float)j * (diffZ * 2.0f + space);
			for(int i=0;i<stackSize;i++) 
			{
				pos[0] = offsetX + (float)i * (diffX * 2.0f + space);				
				trans.setOrigin(offsetPosition + pos);
				btRigidBody *pBody = CreateRigidBody(mass, trans, collisionShape);				
				world->addRigidBody(pBody);				
			}			
		}
		offsetX += diffX;
		offsetZ += diffZ;
		pos[1] += (diffY * 2.0f + space);
		stackSize--;
	}

	return collisionShape;
}

btBvhTriangleMeshShape *CreateBvhTriangleMeshShape(LPD3DXMESH pMesh, TriMeshData *pData)
{
	DWORD numVertices		= pMesh->GetNumVertices();
	DWORD numFaces			= pMesh->GetNumFaces();

	Vertex *v = 0;
	pMesh->LockVertexBuffer(0, (void**)&v);

	// Extract vertices
	pData->vertices = new btScalar[numVertices * 3];	

	for(DWORD i = 0; i < numVertices; i++)
	{
		pData->vertices[i*3+0] = v[i].position.x;
		pData->vertices[i*3+1] = v[i].position.y;
		pData->vertices[i*3+2] = v[i].position.z;		
	}	

	pMesh->UnlockVertexBuffer();

	// Extract indices
	pData->indices = new int[numFaces * 3];	
	WORD* ind = 0;
	pMesh->LockIndexBuffer(0,(void**)&ind);		
	
	//memcpy( &indices, &ind, sizeof(ind));	
	for(DWORD i = 0; i < numFaces; i++)
	{	
		pData->indices[i*3+0] = ind[i*3+0];
		pData->indices[i*3+1] = ind[i*3+1];
		pData->indices[i*3+2] = ind[i*3+2];		
	}

	pMesh->UnlockIndexBuffer();		

	int indexStride = 3 * sizeof(int);
	int vertStride = sizeof(btVector3);

	pData->indexVertexArrays = new btTriangleIndexVertexArray(numFaces, pData->indices, indexStride,
		numVertices, (btScalar*) &pData->vertices[0], sizeof(btScalar) * 3);

	bool useQuantizedAabbCompression = true;
	btBvhTriangleMeshShape *shape = new btBvhTriangleMeshShape(pData->indexVertexArrays, true);	
		
	// The indices, vertices, and array needs to be manually deleted else Bullet will get a access violation
	//delete indices;
	//delete vertices;	
	//delete indexVertexArrays;	

	return shape;
}

btConvexHullShape *CreateConvexHullShape( LPD3DXMESH pMesh )
{
	TriMeshData trimeshshape;
	btBvhTriangleMeshShape *triangleMeshShape = CreateBvhTriangleMeshShape( pMesh, &trimeshshape );						

	btTriangleMesh* triangleMesh = new btTriangleMesh();
	for(DWORD i = 0; i < pMesh->GetNumFaces(); i++)
	{
		int index0 = trimeshshape.indices[i*3];
		int index1 = trimeshshape.indices[i*3+1];
		int index2 = trimeshshape.indices[i*3+2];

		btVector3 vertex0(trimeshshape.vertices[index0*3], trimeshshape.vertices[index0*3+1],trimeshshape.vertices[index0*3+2]);
		btVector3 vertex1(trimeshshape.vertices[index1*3], trimeshshape.vertices[index1*3+1],trimeshshape.vertices[index1*3+2]);
		btVector3 vertex2(trimeshshape.vertices[index2*3], trimeshshape.vertices[index2*3+1],trimeshshape.vertices[index2*3+2]);

		triangleMesh->addTriangle(vertex0,vertex1,vertex2);
	}

	btConvexShape* tmpConvexShape = new btConvexTriangleMeshShape(triangleMesh);

	// Create a hull approximation
	btShapeHull* hull = new btShapeHull(tmpConvexShape);
	btScalar margin = tmpConvexShape->getMargin();
	hull->buildHull(margin);	
	
	btConvexHullShape* simplifiedConvexShape = new btConvexHullShape();
	for(int i= 0; i < hull->numVertices(); i++ )
	{
		simplifiedConvexShape->addPoint(hull->getVertexPointer()[i]);	
	}		

	delete triangleMeshShape;
	delete triangleMesh;
	delete tmpConvexShape;
	delete hull;

	return simplifiedConvexShape;
}

btRigidBody* CreateRigidBody(float mass, const btTransform& startTransform, btCollisionShape* shape)
{
	//rigidbody is dynamic if and only if mass is non zero, otherwise static
	bool isDynamic = (mass != 0.f);

	btVector3 localInertia(0,0,0);
	if (isDynamic)
		shape->calculateLocalInertia(mass,localInertia);
	
	//using motionstate is recommended, it provides interpolation capabilities, and only synchronizes 'active' objects
	btDefaultMotionState* myMotionState = new btDefaultMotionState( startTransform );
	btRigidBody::btRigidBodyConstructionInfo cInfo(mass,myMotionState,shape,localInertia);
	btRigidBody* body = new btRigidBody(cInfo);

	return body;
}

